--------------------------------------------------------------------------------
--   ____  ____
--  /   /\/   /
-- /___/  \  /   Vendor: Xilinx
-- \   \   \/    Version: 1.0
--  \   \        Filename: $RCSfile: global_types_pkg.vhd,v $           
--  /   /        Date Last Modified: $Date: 2008/08/14 13:59:46 $ 
-- /___/   /\    Date Created: April 2008 
-- \   \  /  \   
--  \___\/\___\
-- 
--
-------------------------------------------------------------------------------- 
-- DISCLAIMER OF LIABILITY
--                                 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you 
-- a license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as is" solely for use in developing programs and 
-- solutions for Xilinx devices. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose. 
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications are
-- expressly prohibited.
-- 
--
-- Copyright (c) 2008 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- File       : awgn_gen.vhd
-- Company    : Xilinx
-- Last update: 2003/03/28
-- Platform   : PC
-------------------------------------------------------------------------------
-- Description: This is a AWGN genarator. It accepts the I and Q data bits
--              (QPSK case, for BPSK just ignore the Q data bit), maps it to
--              the Euclidean space, adds AWGN to it and quantize it to widthd
--              data bits. This model puts the pdf of the channel centered
--              arround + ((2**(widthd - 1)) - 1) and
--                      - ((2**(widthd - 1)) - 1)
--              The quantized output data is in two's complement format.
--              I/O description:
--              SNR    : Signal to noise ratio in dB.
--              Rate   : Code Rate.
--              i_hard : The I channel hard data bit.
--              q_hard : The Q channel hard data bit.
--              ce     : Clock enable.
--              reset  : Synchronous reset.
--              clk    : Data clock.
--              i_real : I channel noisy data before quantization.
--              q_real : Q channel noisy data before quantization.
--              i_soft : I channel quantized noisy data in 2's complement
--                       representation.
--              q_soft : Q channel quantized noisy data in 2's complement
--                       representation.
--              widthd : The number of quantized bits.
--              Note   :
--                       Variables SEED1 and SEED2 are initialized to 11 and
--                       31. By changing these vaules, different seeds of the
--                       awgn can be generated.
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.math_real.all;


library work;
use work.global_types_pkg.all;

entity awgn_gen is
  generic (
           output_rate     : integer :=2;
           widthd          : integer := 3;
           snr             : real;
           no_noise        : boolean;
           c_has_erased    : integer := 0;
           c_soft_code     : integer := 0 
           );
  port( ch_hard   : in  std_logic_vector(output_rate-1 downto 0);
        erased    : in  std_logic_vector(output_rate -1 downto 0);
        ce        : in  std_logic;
        reset     : in  std_logic;
        clk       : in  std_logic;
        ch_soft   : out viterbi_array
        );
end awgn_gen;

architecture behav of awgn_gen is
  type  ch_r is array (0 to 7-1) of real;
  type  ch_i is array (0 to output_rate-1) of integer;

  signal ebno     : real := 1.0;
  signal ampl     : real := 1.0;
  signal qlevel   : real := 1.0;
  signal qlevelm1 : real := 1.0;
  signal qleveln  : real := 1.0;
  signal scal     : real := 1.0;
  
  signal ch_tx     : ch_r;
  signal ch_noise  : ch_r;
  signal ch_rcv    : ch_r;
  signal ch_scaled : ch_r;

  constant del : integer := 3;  
  constant rate : real := 0.5;
  signal ch_hard_del : std_logic_vector(output_rate -1 downto 0) := (others => '0');
  type  ch_array is array (0 to output_rate-1) of 
                    std_logic_vector(widthd-1 downto 0); 
  type  ch_del_array is array (0 to output_rate-1) of 
                    std_logic_vector(del-1 downto 0); 
 
  signal ch_soft_int : ch_array;
  
begin

  ampgen : process (ebno, ampl, qlevel)
  begin
    ebno     <= (10.0**(snr/10.0));
    ampl     <= sqrt(2.0*ebno*rate);
    if(widthd > 1) then
       qlevel <= 2.0**(widthd - 1);  
    else
       qlevel <= 2.0;
    end if;   
    qlevelm1 <= qlevel - 1.0;
    qleveln  <= -1.0*qlevelm1;
    scal     <= qlevelm1/ampl;
  end process ampgen;

  txgen : process (clk, reset)
  variable at_start : boolean := true;
  begin
   if(at_start) then
      at_start := false; 
      for i in 0 to output_rate -1 loop 
        ch_tx(i)   <= 1.0;
      end loop;   
   end if;
    if (clk = '1' and clk'event ) then
      if (reset = '1') then
        for i in 0 to output_rate -1 loop 
           ch_tx(i)   <= 0.0;
        end loop;   
      elsif (ce = '1') then
        for i in 0 to output_rate -1 loop 
           if (ch_hard(i) = '1') then
            ch_tx(i) <= -1.0*ampl;
           else
            ch_tx(i) <= ampl;
           end if;
        end loop;   
      end if;
    end if;
  end process txgen;

  noisgen : process (clk, reset)
    variable seed1 : integer := 11;
    variable seed2 : integer := 31;
    variable u1    : real;
    variable u2    : real;
    variable at_start : boolean := true;
    variable cnt : integer := 0;
  begin
   if(at_start) then
      at_start := false; 
      for i in 0 to output_rate -1 loop 
        ch_noise(i)   <= 1.0;
      end loop;   
   end if;
    if (clk = '1' and clk'event ) then
      if (reset = '1' ) then
        for i in 0 to output_rate -1 loop 
         ch_noise(i) <= 0.0;
        end loop; 
      elsif (ce = '1') then
        for i in 0 to output_rate -1 loop 
           uniform(seed1, seed2, u1);
           uniform(seed1, seed2, u2);
           ch_noise(i) <= sqrt(-2.0*log(u1))*cos(2.0*math_pi*u2);
        end loop; 
      end if;
    end if;
  end process noisgen;

  rcvgen : process (clk, reset)
  variable at_start : boolean := true;
  begin
   if(at_start) then
      at_start := false; 
      for i in 0 to output_rate -1 loop 
        ch_rcv(i)   <= 1.0;
      end loop;   
   end if;
    if (clk = '1' and clk'event ) then
      if (reset = '1') then
        for i in 0 to output_rate -1 loop 
           ch_rcv(i) <= 0.0;
        end loop; 
      elsif (ce = '1') then
        for i in 0 to output_rate -1 loop 
          if(no_noise) then
            ch_rcv(i) <= ch_tx(i);
          else
            ch_rcv(i) <= ch_tx(i) + ch_noise(i);
          end if;
        end loop; 
      end if;
    end if;
  end process rcvgen;

  scale : process (ch_rcv, scal)
  begin
    for i in 0 to output_rate -1 loop 
       ch_scaled(i) <= scal*ch_rcv(i);
    end loop;
  end process scale;


  quant : process (clk, reset)
    variable ch_temp : ch_r;
    variable ch_del : ch_del_array;
    variable ch_conv : ch_i;
    constant convert : integer := (2**(widthd-1)+1);
    constant convert1 : integer := (2**(widthd-1)-1);
    variable at_start : boolean := true;
  begin
   if(at_start) then
      at_start := false; 
      for i in 0 to output_rate -1 loop 
        ch_del(i)   := (others => '0');
        if(c_soft_code = soft_signed_magnitude ) then
          ch_soft_int(i) <= (others => '0');
        else
          ch_soft_int(i)(widthd-1) <= '0'; -- to avoid timing conflicts at start
          ch_soft_int(i)(widthd-2 downto 0) <= (others => '1');
        end if;
      end loop;   
   end if;
    if (clk = '1' and clk'event ) then
      if (reset = '1' ) then
        for i in 0 to output_rate -1 loop 
          ch_del(i) := (others => '0');
          ch_soft_int(i) <= (others => '0');
          if(c_soft_code = soft_signed_magnitude ) then
            ch_soft_int(i) <= (others => '0');
          else
            ch_soft_int(i)(widthd-1) <= '0'; -- to avoid timing conflicts at start
            ch_soft_int(i)(widthd-2 downto 0) <= (others => '1');
          end if;
        end loop;
      elsif (ce = '1') then
        for i in 0 to output_rate -1 loop 
           ch_del(i)(del-1 downto 1) := ch_del(i)(del-2 downto 0); 
           ch_del(i)(0) :=  ch_hard(i);
        
           if (ch_scaled(i) > qlevelm1) then
             ch_temp(i) := qlevelm1;
           elsif (ch_scaled(i) <= qleveln) then
             ch_temp(i) := qleveln - 1.0;
           else
             ch_temp(i) := ch_scaled(i);
           end if;
           ch_hard_del(i) <= ch_del(i)(del-1);
           -- convert integers to correct format
           ch_conv(i) := integer(floor(ch_temp(i)));
           if(widthd = 1) then
            -- hard coding
              if(ch_conv(i) < 0) then
                  ch_conv(i) := 1;
              else
                  ch_conv(i) := 0;
              end if; 
           else
              if( c_soft_code = soft_signed_magnitude ) then
                 if(ch_conv(i) < 0) then
                    ch_conv(i) := -ch_conv(i) - convert;
                 end if;       
              else
                 if(ch_conv(i) < 0) then
                    ch_conv(i) := -ch_conv(i) - convert;
                 end if;       
                 if(ch_conv(i) >= 0) then
                    ch_conv(i) := -ch_conv(i) + convert1;
                 end if;       
              end if;
           end if; 
           if(c_has_erased /= 0) then 
              if(erased(i) = '1') then
               ch_conv(i) := 0;
              end if;
           end if;
           ch_soft_int(i) <= to_vector(widthd, ch_conv(i));
        end loop;   
      end if;
    end if;
  end process quant;

  gen_out:
  for i in 0 to output_rate - 1 generate
     ch_soft(i)(widthd-1 downto 0) <= ch_soft_int(i)(widthd-1 downto 0);
 end generate;   
end behav;
