--------------------------------------------------------------------------------
--   ____  ____
--  /   /\/   /
-- /___/  \  /   Vendor: Xilinx
-- \   \   \/    Version: 1.0
--  \   \        Filename: conv_enc.vhd          
--  /   /        
-- /___/   /\    Date Created: April 2009 
-- \   \  /  \   
--  \___\/\___\
-- 
--
-------------------------------------------------------------------------------- 
--Legal Disclaimer
--
--(c) Copyright 2002 - 2009 Xilinx, Inc. All rights reserved. 
--
--This file contains confidential and proprietary information
--of Xilinx, Inc. and is protected under U.S. and
--international copyright and other intellectual property
--1aws.
----
--DISCLAIMER
--This disclaimer is not a license and does not grant any
--rights to the materials distributed herewith. Except as
--otherwise provided in a valid license issued to you by
--Xilinx, and to the maximum extent permitted by applicable
--law: (1) THESE MATERIALS ARE MADE AVAILABLE "AS IS" AND
--WITH ALL FAULTS, AND XILINX HEREBY DISCLAIMS ALL WARRANTIES
--AND CONDITIONS, EXPRESS, IMPLIED, OR STATUTORY, INCLUDING
--BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, NON-
--INFRINGEMENT, OR FITNESS FOR ANY PARTICULAR PURPOSE; and
--(2) Xilinx shall not be liable (whether in contract or tort,
--including negligence, or under any other theory of
--liability) for any loss or damage of any kind or nature
--related to, arising under or in connection with these
--materials, including for any direct, or any indirect,
--special, incidental, or consequential loss or damage
--(including loss of data, profits, goodwill, or any type of
--loss or damage suffered as a result of any action brought
--by a third party) even if such damage or loss was
--reasonably foreseeable or Xilinx had been advised of the
--possibility of the same.
----
--CRITICAL APPLICATIONS
--Xilinx products are not designed or intended to be fail-
--safe, or for use in any application requiring fail-safe
--performance, such as life-support or safety devices or
--systems, Class III medical devices, nuclear facilities,
--applications related to the deployment of airbags, or any
--other applications that could lead to death, personal
--injury, or severe property or environmental damage
--individually and collectively, "Critical
--Applications"). Customer assumes the sole risk and
--liability of any use of Xilinx products in Critical
--Applications, subject only to applicable laws and
--regulations governing limitations on product liability.
----
--THIS COPYRIGHT NOTICE AND DISCLAIMER MUST BE RETAINED AS
--PART OF THIS FILE AT ALL TIMES.
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--
-- Function : convolution encoder with tail biting 
-----------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.global_types_pkg.all;


entity conv_enc is
  port
  (
	  --inputs
    clk        : in std_logic;
	  ext2all    : in GLB_InterfaceEXT2ALL     := GLB_InterfaceEXT2ALL_DEFAULT;
	  puip       : in GLB_Data                 := GLB_Data_DEFAULT;
    cts        : in std_logic;
	  --outputs
    rfd        : out std_logic;
	  puop       : out GLB_Data                := GLB_Data_DEFAULT
   );
end conv_enc;

architecture synth of conv_enc is

constant CB_ADDR_W : INTEGER := 16; 
constant CONSTRAINT_LENGTH : integer := 7;
constant CONSTRAINT_LENGTH_M1 : integer := CONSTRAINT_LENGTH -1;

signal puip_cb,puop_cb : GLB_Data;
signal puop_enc :  GLB_Data;
-- circular buffer signals
signal cb_rfd,valid_del : std_logic;
-- encoding registers
signal d0_reg,d1_reg,d2_reg : std_logic_vector(CONSTRAINT_LENGTH-1 downto 0);
signal start,valid_and : std_logic;

begin

	------------------------------------------------------------------------------
  -- circular buffer to take input data to handle the tail biting bits
  -- instantiate the circular buffer
  -- data is input sequentially and read out first 6 tail bits followed
  -- by rest of block to be encoded
	------------------------------------------------------------------------------
  -- sort out input data to buffer, using puip.data to pass in sys as a vector
  -- and if we can't start a new block stop the flow
  puip_cb.valid <= puip.valid and cb_rfd;
  puip_cb.data(0) <= puip.data(0);
  puip_cb.cb_start <= puip.cb_start and puip.valid and cb_rfd;
  puip_cb.cb_done <= puip.cb_done and puip.valid and cb_rfd;
  

  cb: entity work.circ_buffer_enc
    generic map
    (
      c_addr_width => CB_ADDR_W,
      c_data_width => 1
    )
    port map
    (
	    -- data input handling
      data_in     => puip_cb.data(0 downto 0), -- data written to buffer sequentially
      fd_wr       => puip_cb.cb_start,-- first data start of buffer block -- only valid when nd and rfd are high
      ld_wr       => puip_cb.cb_done,-- last data end of buffer block -- only valid when nd and rfd are high
      nd          => puip_cb.valid,-- input data is valid if rfd is high
      rfd         => cb_rfd,--  data can be written
      -- data output handling
      cts         => cts,
      data_out    => puop_cb.data(0 downto 0),
      rdy         => puop_cb.valid,
      start       => puop_cb.cb_start,
      done        => puop_cb.cb_done,
      -- clocks etc
      clk         => clk,
	    ext2all     => ext2all
     );
  


	------------------------------------------------------------------------------
 	-- encode the data
  -- take the data from the circular buffer
  -- this is the last 6 bits of the packet followed by the packet
  -- first six bits of output of the encoder for each block are not valid
  -- bits 1 to 6 are the constraint register 0 is an initial register on the output
  -- from the buffer
  -- initial register on data going into encoder
  -- fixed encoder based on the 3gpp LTE standard constraint length 7 output rate3
  -- G0 = 133 octal G1 = 171 octal G2 = 165 octal
	------------------------------------------------------------------------------
	enc: process(clk)
	begin
		if rising_edge(clk) and ext2all.ce='1' then
      if (ext2all.reset = '1') then
        puop_enc <= GLB_Data_DEFAULT;
        d0_reg <= (others => '0');
        d1_reg <= (others => '0');
        d2_reg <= (others => '0');
        valid_del <= '0';
        start <= '0';
        valid_and <= '0';
      else
        if (cts = '1') then
          if (puop_cb.valid = '1') then 
            -- convolutional encoding data moving through the six bit register
            d0_reg(0) <= puop_cb.data(0);
            d1_reg(0) <= puop_cb.data(0);
            d2_reg(0) <= puop_cb.data(0);
            d0_reg(CONSTRAINT_LENGTH-1 downto 1) <= d0_reg(CONSTRAINT_LENGTH-2 downto 0);
            d1_reg(CONSTRAINT_LENGTH-1 downto 1) <= d1_reg(CONSTRAINT_LENGTH-2 downto 0);
            d2_reg(CONSTRAINT_LENGTH-1 downto 1) <= d2_reg(CONSTRAINT_LENGTH-2 downto 0);
          end if;  
          puop_enc.data(0) <= d0_reg(0) xor d0_reg(2) xor d0_reg(3) xor d0_reg(5) xor d0_reg(6);
          puop_enc.data(1) <= d1_reg(0)  xor d1_reg(1)  xor d1_reg(2)  xor d1_reg(3)  xor d1_reg(6);
          puop_enc.data(2) <= d2_reg(0)  xor d2_reg(1)  xor d2_reg(2)  xor d2_reg(4)  xor d2_reg(6);
          if(puop_cb.cb_start = '1') then
            valid_and <= '1';
          elsif (puop_cb.cb_done = '1') then 
            valid_and <= '0';
          end if;
          valid_del <= puop_cb.valid;
          puop_enc.valid <= valid_del and valid_and;

          start <= puop_cb.cb_start; -- start from the buffer is one early
          puop_enc.cb_start <= start;

          puop_enc.cb_done <= puop_cb.cb_done;
          
        end if;-- cts
      end if;-- reset
		end if; -- clk
	end process enc;


	------------------------------------------------------------------------------
  -- pass out data
	------------------------------------------------------------------------------
  PUOP.data <= puop_enc.data;
  PUOP.valid <= puop_enc.valid;
  PUOP.cb_start <= puop_enc.cb_start;
  PUOP.cb_done <= puop_enc.cb_done;
  RFD <= cb_rfd;
  

end synth;

------------ EOF  ------------
