--------------------------------------------------------------------------------
--   ____  ____
--  /   /\/   /
-- /___/  \  /   Vendor: Xilinx
-- \   \   \/    Version: 1.0
--  \   \        Filename: $RCSfile: global_types_pkg.vhd,v $           
--  /   /        Date Last Modified: $Date: 2008/08/14 13:59:46 $ 
-- /___/   /\    Date Created: April 2008 
-- \   \  /  \   
--  \___\/\___\
-- 
--
-------------------------------------------------------------------------------- 
-- DISCLAIMER OF LIABILITY
--                                 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you 
-- a license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as is" solely for use in developing programs and 
-- solutions for Xilinx devices. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose. 
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications are
-- expressly prohibited.
-- 
--
-- Copyright (c) 2008 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--------------------------------------------------------------------------------
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;


package global_types_pkg IS

-------------------------------------------  
-- Viterbi constants
-- New model required if these generics are
-- changed
-- fixed encoder based on the 3gpp LTE standard 
-- constraint length 7 output rate3
-- G0 = 133 octal G1 = 171 octal G2 = 165 octal

-------------------------------------------  
  type viterbi_array      is array (0 to 2) of std_logic_vector(4 downto 0); 
  constant soft_signed_magnitude   : integer := 0;
  constant soft_offset_binary      : integer := 1;
  constant soft_signed             : integer := 2;
  
  constant CONSTRAINT_LENGTH : positive := 7;
  constant SOFT_W            : positive := 4;
  constant OUTPUT_RATE       : positive := 3;
  constant TRACEBACK_LENGTH  : positive := 42;
  constant TB_W              : positive := 6;
  constant SOFT_CODE         : integer  := soft_signed_magnitude;


-------------------------------------------  
-- FIFO records
-------------------------------------------  

  constant FIFO_DEPTH_W : positive := 4; -- overflow handled through the ctrl_rfd signal
  constant FIFO_DEPTH : positive := 2**FIFO_DEPTH_W; -- overflow handled through the ctrl_rfd signal
  constant MAX_FIFO_W : positive := 256;  
  
  type fifo_r is record
   wr_data    : std_logic_vector(MAX_FIFO_W-1 downto 0);
   wr_enable  : std_logic;
   rd_data    : std_logic_vector(MAX_FIFO_W-1 downto 0);
   rd_enable  : std_logic;
   rd_valid   : std_logic;
   full       : std_logic;
   empty      : std_logic;
  end record;  
  
  constant fifo_r_DEFAULT : fifo_r := (
    wr_data       => (others =>'0'),
    wr_enable     => '0',
    rd_data       => (others =>'0'),
    rd_enable     => '0',
    rd_valid      => '0',
    full          => '0',
    empty         => '0'
  );


-----------------------------------------------------------------------------------------------------------------------
--Define the common interface from the External System (EXT) to all sub-systems
	type GLB_InterfaceEXT2ALL is record
		ce:             std_logic;              --System clock enable
		reset:          std_logic;              --Active to start core initialisation
		--                                      --  Note that this signal has higher precedence than ce
		--                                      --  i.e. if core_init='1' and ce='0' the core will start initialisation
	end record;

  
	constant GLB_InterfaceEXT2ALL_DEFAULT: GLB_InterfaceEXT2ALL:=
  (
		ce           =>'1',  --Clock enable defaults to '1'
		reset        =>'0'
	);

-----------------------------------------------------------------------------------------------------------------------
-- input/output data busses
  type GLB_Data is record
    data:       std_logic_vector(SOFT_W*OUTPUT_RATE-1 downto 0); 
    valid:      std_logic;   -- data is valid
    cb_start:   std_logic;   -- start of block
    cb_done:    std_logic;   -- end of block
  end record;

  constant GLB_Data_DEFAULT  : GLB_Data :=
  (
    data       => "000000000000",
    valid      => '0',
    cb_start   => '0',
    cb_done    => '0'
  );
  
  function to_uint (a : std_logic_vector) return integer;

  function to_vector (size : integer; val : integer) return std_logic_vector;

end global_types_pkg;


------------------------------------------------------------------------------------------------------------------------
package body global_types_pkg is

-------------------------------------------------------------------------------
-- Convert a std_logic_vector to an unsigned integer
-------------------------------------------------------------------------------
  function to_uint (a : std_logic_vector) return integer is

    alias av     : std_logic_vector (1 to a'length) is a;
    variable val : integer := 0;
    variable b   : integer := 1;

  begin

    for i in a'length downto 1 loop

      if (av(i) = '1') then             -- if LSB is '1',
        val := val + b;                 -- add value for current bit position
      end if;

      b := b * 2;                       -- Shift left 1 bit

    end loop;
    return val;

  end to_uint;

-------------------------------------------------------------------------------
-- Convert an integer to a std_logic_vector
-------------------------------------------------------------------------------
  function to_vector (size : integer; val : integer) return std_logic_vector is

    variable vec      : std_logic_vector (1 to size);
    variable a        : integer;
    variable negative : boolean := false;

  begin

    negative := (val < 0);
    a        := val;

    for i in size downto 1 loop
      if ((a mod 2) = 1) then
        vec(i) := '1';
      else
        vec(i) := '0';
      end if;

      a := a / 2;

    end loop;

    if negative then
      vec := not(vec) + '1';
    end if;

    return vec;

  end to_vector;



end global_types_pkg;
